/* interp.c    interpreter i-code	9 Jan 97

	start project RZ				3 Apr 2001
	project A1						23 Oct 2002
	som-c change to dynamic array  	4 Jan 2004
	rewrite using big switch	  	13 Jan 2004
	new call/ret					22 Mar 2004
	add icSys 19 loadfile in som3	9 Nov 2004
	eliminate tail call				Feb 2005
	som-v16 add file i/o 			7 Nov 2005
	t-code som-v17					1 Jan 2006 (Happy new year!)

*/

#include "compile.h"

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..63 locals, 64..255 globals
*/

int M[MAXMEM];				// data and code segment
int SS[MAXSS];				// stack segment
int *CS = &M[MAXDS];		// code segment
int DS = MAXSYS;			// current free data segment

PRIVATE FILE *filep[MAXFILE]; // store file pointer

PRIVATE int fp = 1;			// current state of computation
PRIVATE int opcode,data;	// current instruction

void initfile(void){
	int i;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
}
// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}
/*
void dumpDS(void){
	int i;
	for(i=1; i<=10; i++)
		printf("%d:%d ",i,M[i]);
	nl();
	for(i=64; i<70; i++)
		printf("%d:%d ",i,M[i]);
	printf("\n");
}

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
	prCode(stdout,ip);
	nl();
	dumpDS();
	printf("fp:%d ",fp);
	a = fp - 10;
	if( a < 0 ) a = 0;
	for(i=a; i<=fp+5; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}
*/

PRIVATE void error(char *mess, int i){
	printf("%s at ip %d\n",mess,i);
	exit(0);
}

// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		seterror("alloc: out of memory");
	DS += a;
	return b;
}
// exec syscall n, optional in: a, b
PRIVATE void xsys(int a, int b, int n){
	int f;
	char name[NAMELEN];

	switch(n){
	case 1: printf("%d",a); break;		// pr int
	case 2: printf("%c",a); break;		// pr char
	case 3: M[RETVAL]=getchar(); break;	// getchar
//	case 4: break;						// gets
	case 5:								// fopen fname mode
		strunpack(name,a);				// file name
		f = newfp();
		if(f == 0) seterror("cannot open, too many files");
		switch(b){
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"rw"); break;
		default: seterror("unknown file mode");
		}
		M[RETVAL] = f;				// return file pointer
		break;
	case 6:							// fclose fp
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:	fprintf(filep[b],"%d",a); break; // fpr fp num
	case 8:	fprintf(filep[b],"%c",a); break; // fpr fp c
	case 9:							// fgetc fp
		b = fgetc(filep[a]);
		M[RETVAL] = b;
		break;
	case 10: M[RETVAL] = xalloc(a); break;	// 10
	case 11:				// load
		strunpack(name,a); 	// file name in som-string
		loadfile(name);
		break;

	default: seterror("undefined syscall");
	}
}

// decoder
#define dec3(d,a,b,c)	a=(d)&255; b=(d>>8)&255; c=(d>>16)&255
#define dec2(d,a,b)		a=(d)&255; b=(d>>8)&255
#define dec1(d,a)		a=(d)&255
#define decd(d,a,n)		a=(d)&255; n=(d)>>8

void eval(int ref){
//	int cnt = 0;
	int runflag, ip;
	int i, low, hi, arity, nlocal;
	int a1, a2, a3, c, d;

	ip = ref;
	runflag = 1;
    while(runflag ) {
//		if( cnt++ > 5000 ) break;

		c = CS[ip];
		opcode = c & 255;
		data = c >> 8;

// ---- start exec -----

	switch(opcode){
	case tLit:
	case tAds:
		M[RETVAL] = data;
		ip++;
		break;
	case tLdx:
		dec3(data,a1,a2,a3);
		M[a1] = M[ M[a2] + M[a3] ];
		ip++;
		break;
	case tStx:
		dec3(data,a1,a2,a3);
		M[ M[a2] + M[a3] ] = M[a1];
		ip++;
		break;
	case tJmp:
		ip = data;
		break;
	case tJf:
		decd(data,a1,d);
		if( M[a1] == FALSE) ip += d;
		else ip++;
		break;
	case tJt:
		decd(data,a1,d);
		if( M[a1] != FALSE) ip += d;
		else ip++;
		break;
	case tCall:					// fun arity nlocal 0
		c = CS[data];
		arity = (c >> 8) & 255;		// a1
		nlocal = (c >> 16) & 255;	// a2
		if (fp + nlocal + FUDGE > MAXSS)
			error("stack overflow", ip);

		fp -= arity;			// fp before 1st param
		for(i=1; i<=nlocal; i++)
			if(i<=arity){
				d = SS[fp+i];
				SS[fp+i] = M[i]; // save reg
				M[i] = d;		// pass param
			}else
				SS[fp+i] = M[i]; // save reg

		SS[fp+i] = ip+1;		// ret ads
		fp += nlocal+1;
		ip = data+1;
		break;
	case tRet:					// ret v nlocal 0
		dec2(data,a1,a2);
		ip = SS[fp];			// restore ip
		M[RETVAL] = M[a1];		// ret value
		fp -= a2+1;				// delete AR
		for(i=1; i<=a2; i++)	// restore reg
			M[i] = SS[fp+i];
		break;
	case tCallt:
		c = CS[data];
		arity = (c >> 8) & 255; // a1
		fp -= arity;			// fp before 1st param
		for(i=1; i<=arity; i++)
			M[i] = SS[fp+i];
		ip = data+1;
		break;
	case tCallf:				// refer to sym tab
		arity = getArg(data);
		nlocal = arity+getLv(data);
		if (fp + nlocal + FUDGE > MAXSS)
			error("stack overflow", ip);

		fp -= arity;			// fp before 1st param
		for(i=1; i<=nlocal; i++)
			if(i<=arity){
				d = SS[fp+i];
				SS[fp+i] = M[i]; // save reg
				M[i] = d;		// pass param
			}else
				SS[fp+i] = M[i]; // save reg

		SS[fp+i] = ip+1;		// ret ads
		fp += nlocal+1;
		ip = getRef(data)+1;
		break;
	case tAdd:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] + M[a3];
		ip++;
		break;
	case tSub:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] - M[a3];
		ip++;
		break;
	case tMul:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] * M[a3];
		ip++;
		break;
	case tDiv:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] / M[a3];
		ip++;
		break;
	case tMod:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] % M[a3];
		ip++;
		break;
	case tAnd:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] & M[a3];
		ip++;
		break;
	case tOr:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] | M[a3];
		ip++;
		break;
	case tXor:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] ^ M[a3];
		ip++;
		break;
	case tShl:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] << M[a3];
		ip++;
		break;
	case tShr:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] >> M[a3];
		ip++;
		break;
	case tEq:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] == M[a3];
		ip++;
		break;
	case tNe:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] != M[a3];
		ip++;
		break;
	case tLt:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] < M[a3];
		ip++;
		break;
	case tLe:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] <= M[a3];
		ip++;
		break;
	case tGt:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] > M[a3];
		ip++;
		break;
	case tGe:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] >= M[a3];
		ip++;
		break;
	// immediate
	case tAddi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] + a3;
		ip++;
		break;
	case tSubi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] - a3;
		ip++;
		break;
	case tMuli:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] * a3;
		ip++;
		break;
	case tDivi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] / a3;
		ip++;
		break;
	case tModi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] % a3;
		ip++;
		break;
	case tAndi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] & a3;
		ip++;
		break;
	case tOri:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] | a3;
		ip++;
		break;
	case tXori:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] ^ a3;
		ip++;
		break;
	case tShli:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] << a3;
		ip++;
		break;
	case tShri:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] >> a3;
		ip++;
		break;
	case tEqi:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] == a3;
		ip++;
		break;
	case tNei:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] != a3;
		ip++;
		break;
	case tLti:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] < a3;
		ip++;
		break;
	case tLei:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] <= a3;
		ip++;
		break;
	case tGti:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] > a3;
		ip++;
		break;
	case tGei:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] >= a3;
		ip++;
		break;
	case tNot:
		dec3(data,a1,a2,a3);
		M[a1] = M[a2] == FALSE;
		ip++;
		break;
	// special
	case tMov:
		dec2(data,a1,a2);
		M[a1] = M[a2];
		ip++;
		break;
	case tMovi:
		decd(data,a1,d);
		M[a1] = d;
		ip++;
		break;
	case tMovd:
		dec2(data,a1,a2);
		M[a1] = M[ M[a2] ];
		ip++;
		break;
	case tPass:
		dec2(data,a1,a2);
		fp++;
		SS[fp] = M[a2];
		ip++;
		break;
	case tPassi:
		decd(data,a1,d);
		fp++;
		SS[fp] = d;
		ip++;
		break;
	case tEfor:
		decd(data,a1,d);
		M[a1]++;
		if(M[a1] <= M[a1+1])	// adj(a1)
			ip += d;
		else
			ip++;
		break;
	case tCase:
		decd(data,a1,d);
		low = CS[ip+1] >> 8;
		hi = CS[ip+2] >> 8;
		c = M[a1];
		if( c < low || c > hi )
			ip += d;
		else
			ip += c-low+3;
		break;
	case tSys:
		dec3(data,a1,a2,a3);
		xsys(M[a1],M[a2],a3);
		ip++;
		break;
	case tEnd:
		runflag = 0;
		break;
	default:
		error("undefined opcode", ip);
	}
// ------ end exec --------

    }
//	printf("noi %d\n",cnt);
}



